from typing import Any, Dict

labels: Dict[str, Any] = {
    "label_hash": "1b4398aa-1d4d-4873-8470-2ac9fa373ac0",
    "branch_name": "main",
    "created_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "last_edited_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "data_hash": "aaa6bc82-9f89-4545-adbb-f271bf28cf99",
    "annotation_task_status": "QUEUED",
    "is_shadow_data": False,
    "dataset_hash": "d947e847-88c8-4762-aee5-53dc7eeab852",
    "dataset_title": "Tracking bug dataset",
    "data_title": "failing_video_new.mp4",
    "data_type": "video",
    "data_units": {
        "cd63592b-a9e4-47d7-8563-aaed75c5afdc": {
            "data_hash": "cd63592b-a9e4-47d7-8563-aaed75c5afdc",
            "data_title": "failing_video_new.mp4",
            "data_type": "video/mp4",
            "data_sequence": 0,
            "labels": {
                "0": {
                    "objects": [
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 11 Jul 2022 17:10:51 UTC",
                            "createdBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "428fba2b",
                            "featureHash": "249c9370",
                            "lastEditedAt": "Mon, 09 Jan 2023 08:19:16 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.0986, "w": 0.0859, "x": 0.2964, "y": 0.2516},
                        },
                        {
                            "name": "point",
                            "color": "#68BC00",
                            "point": {"0": {"x": 0.4243, "y": 0.3458}},
                            "shape": "point",
                            "value": "point",
                            "createdAt": "Mon, 11 Jul 2022 17:10:52 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Mon, 11 Jul 2022 17:10:52 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "82724bb2",
                            "featureHash": "d02c6e13",
                            "manualAnnotation": True,
                        },
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 09 Jan 2023 08:18:27 UTC",
                            "createdBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "u8BIOLmY",
                            "featureHash": "249c9370",
                            "lastEditedAt": "Mon, 09 Jan 2023 08:19:10 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.052, "w": 0.0626, "x": 0.3052, "y": 0.4969},
                        },
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 09 Jan 2023 08:18:28 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Mon, 09 Jan 2023 08:18:28 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "/ghwDXVK",
                            "featureHash": "249c9370",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.0914, "w": 0.0986, "x": 0.5313, "y": 0.5914},
                        },
                    ],
                    "classifications": [],
                },
                "1": {
                    "objects": [
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "428fba2b",
                            "featureHash": "249c9370",
                            "manualAnnotation": False,
                            "boundingBox": {"h": 0.0979, "w": 0.0844, "x": 0.2641, "y": 0.2417},
                        },
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 09 Jan 2023 08:19:12 UTC",
                            "createdBy": "denis@cord.tech",
                            "isDeleted": False,
                            "confidence": 1,
                            "objectHash": "u8BIOLmY",
                            "featureHash": "249c9370",
                            "lastEditedAt": "Mon, 09 Jan 2023 08:19:12 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.052, "w": 0.0626, "x": 0.3052, "y": 0.4969},
                        },
                    ],
                    "classifications": [],
                },
                "2": {
                    "objects": [
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "428fba2b",
                            "featureHash": "249c9370",
                            "manualAnnotation": False,
                            "boundingBox": {"h": 0.0979, "w": 0.0844, "x": 0.2641, "y": 0.2375},
                        }
                    ],
                    "classifications": [],
                },
                "3": {
                    "objects": [
                        {
                            "name": "Epiglottis ",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "epiglottis_",
                            "createdAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Mon, 11 Jul 2022 17:11:05 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "428fba2b",
                            "featureHash": "249c9370",
                            "manualAnnotation": False,
                            "boundingBox": {"h": 0.0979, "w": 0.0844, "x": 0.2687, "y": 0.2375},
                        }
                    ],
                    "classifications": [],
                },
            },
            "data_fps": 25.0,
            "data_duration": 153.16,
            "data_link": "cord-videos-dev/lFW59RQ9jcT4vHZeG14m8QWJKug1/cd63592b-a9e4-47d7-8563-aaed75c5afdc",
            "width": 640,
            "height": 480,
        }
    },
    "object_answers": {
        "428fba2b": {"objectHash": "428fba2b", "classifications": []},
        "82724bb2": {"objectHash": "82724bb2", "classifications": []},
        "u8BIOLmY": {"objectHash": "u8BIOLmY", "classifications": []},
        "/ghwDXVK": {"objectHash": "/ghwDXVK", "classifications": []},
    },
    "classification_answers": {},
    "object_actions": {},
    "label_status": "LABELLED",
}

ontology: Dict[str, Any] = {
    "objects": [
        {"id": "1", "name": "Epiglottis ", "color": "#D33115", "shape": "bounding_box", "featureNodeHash": "249c9370"},
        {"id": "2", "name": "Larynx", "color": "#E27300", "shape": "bounding_box", "featureNodeHash": "e3c87d43"},
        {"id": "3", "name": "Oesophagus ", "color": "#16406C", "shape": "bounding_box", "featureNodeHash": "c6eb717a"},
        {"id": "4", "name": "Z-line ", "color": "#FE9200", "shape": "polyline", "featureNodeHash": "8b64f336"},
        {"id": "5", "name": "Stomach", "color": "#FCDC00", "shape": "bounding_box", "featureNodeHash": "f44083a3"},
        {"id": "6", "name": "Antrum ", "color": "#DBDF00", "shape": "bounding_box", "featureNodeHash": "7f3f3ed6"},
        {"id": "7", "name": "Incisura ", "color": "#A4DD00", "shape": "polygon", "featureNodeHash": "8dd583bb"},
        {"id": "8", "name": "D1", "color": "#68CCCA", "shape": "bounding_box", "featureNodeHash": "58bcccb6"},
        {"id": "9", "name": "D2", "color": "#73D8FF", "shape": "bounding_box", "featureNodeHash": "761eff9d"},
        {"id": "10", "name": "LES", "color": "#AEA1FF", "shape": "bounding_box", "featureNodeHash": "c01997bb"},
        {
            "id": "11",
            "name": "Pyloric opening ",
            "color": "#FCC400",
            "shape": "bounding_box",
            "featureNodeHash": "a96af107",
        },
        {"id": "12", "name": "Fundus ", "color": "#B0BC00", "shape": "bounding_box", "featureNodeHash": "f2870195"},
        {"id": "13", "name": "point", "color": "#68BC00", "shape": "point", "featureNodeHash": "d02c6e13"},
    ],
    "classifications": [
        {
            "id": "1",
            "featureNodeHash": "ebbaf6b6",
            "attributes": [
                {
                    "id": "1.1",
                    "name": "Stomach",
                    "type": "radio",
                    "featureNodeHash": "6e23d0b8",
                    "required": False,
                    "dynamic": False,
                    "options": [
                        {"id": "1.1.1", "label": "Fundus", "value": "fundus", "featureNodeHash": "a48fdc54"},
                        {"id": "1.1.2", "label": "Antrum ", "value": "antrum_", "featureNodeHash": "7103db43"},
                    ],
                }
            ],
        },
        {
            "id": "2",
            "featureNodeHash": "7d8b28e8",
            "attributes": [
                {
                    "id": "2.1",
                    "name": "Fundus ",
                    "type": "radio",
                    "featureNodeHash": "a179f3a2",
                    "required": False,
                    "dynamic": False,
                }
            ],
        },
        {
            "id": "3",
            "featureNodeHash": "7f14717c",
            "attributes": [
                {
                    "id": "3.1",
                    "name": "",
                    "type": "radio",
                    "featureNodeHash": "bcb6c6c6",
                    "required": False,
                    "dynamic": False,
                    "options": [
                        {"id": "3.1.1", "label": "Radio One", "value": "radio_one", "featureNodeHash": "b954d23c"},
                        {"id": "3.1.2", "label": "Radio Two", "value": "radio_two", "featureNodeHash": "a6400563"},
                    ],
                }
            ],
        },
    ],
}



"""
Code Block Name: Bounding Boxes Images/Videos
"""

# Import dependencies
from encord import EncordUserClient, Project
from encord.objects import ChecklistAttribute, Object, ObjectInstance, Option, RadioAttribute, TextAttribute
from encord.objects.coordinates import BoundingBoxCoordinates

SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
# SSH_PATH = get_ssh_key() # replace it with ssh key
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
BUNDLE_SIZE = 100

# Create user client using ssh key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Get project for which labels are to be added
project: Project = user_client.get_project(PROJECT_ID)

# Create radio button attribute for cherry type
ontology_structure = project.ontology_structure

# Find a bounding box annotation object in the project ontology
box_ontology_object: Object = ontology_structure.get_child_by_title(title="Cherries", type_=Object)
assert box_ontology_object is not None, "Bounding box object 'Cherries' not found in ontology."

cherry_type_radio_attribute = ontology_structure.get_child_by_title(type_=RadioAttribute, title="Type?")
assert cherry_type_radio_attribute is not None, "Radio attribute 'Type?' not found in ontology."

# Create options for the radio buttons
bing_option = cherry_type_radio_attribute.get_child_by_title(type_=Option, title="Bing")
king_option = cherry_type_radio_attribute.get_child_by_title(type_=Option, title="King")
rainier_option = cherry_type_radio_attribute.get_child_by_title(type_=Option, title="Rainier")
other_cherry_option = cherry_type_radio_attribute.get_child_by_title(type_=Option, title="Other cherry type")

assert all([bing_option, king_option, rainier_option, other_cherry_option]), (
    "One or more cherry type options are missing."
)

# Bing Qualities
bing_checklist_attribute = ontology_structure.get_child_by_title(type_=ChecklistAttribute, title="Bing Qualities?")
assert bing_checklist_attribute is not None, "Checklist attribute 'Bing Qualities?' not found."

bing_plump_option = bing_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
bing_juicy_option = bing_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
bing_large_option = bing_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert all([bing_plump_option, bing_juicy_option, bing_large_option]), "One or more Bing quality options are missing."

# King Qualities
king_checklist_attribute = ontology_structure.get_child_by_title(type_=ChecklistAttribute, title="King Qualities?")
assert king_checklist_attribute is not None, "Checklist attribute 'King Qualities?' not found."

king_plump_option = king_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
king_juicy_option = king_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
king_large_option = king_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert all([king_plump_option, king_juicy_option, king_large_option]), "One or more King quality options are missing."

# Rainier Qualities
rainier_checklist_attribute = ontology_structure.get_child_by_title(
    type_=ChecklistAttribute, title="Rainier Qualities?"
)
assert rainier_checklist_attribute is not None, "Checklist attribute 'Rainier Qualities?' not found."

rainier_plump_option = rainier_checklist_attribute.get_child_by_title(type_=Option, title="Plump")
rainier_juicy_option = rainier_checklist_attribute.get_child_by_title(type_=Option, title="Juicy")
rainier_large_option = rainier_checklist_attribute.get_child_by_title(type_=Option, title="Large")
assert all([rainier_plump_option, rainier_juicy_option, rainier_large_option]), (
    "One or more Rainier quality options are missing."
)

# Other Cherry Types
other_cherry_option_text_attribute = ontology_structure.get_child_by_title(
    type_=TextAttribute, title="Specify cherry type"
)
assert other_cherry_option_text_attribute is not None, "Text attribute 'Specify cherry type' not found in ontology."

# Dictionary of labels per data unit and per frame with cherry type specified, including quality options
video_frame_labels = {
    "cherries-001.jpg": {
        0: {
            "label_ref": "cherry_001",
            "coordinates": BoundingBoxCoordinates(height=0.4, width=0.4, top_left_x=0.0, top_left_y=0.0),
            "cherry_type": "Bing",
            "bing_quality_options": "Plump, Juicy",
        }
    },
    "cherries-010.jpg": {
        0: [
            {
                "label_ref": "cherry_002",
                "coordinates": BoundingBoxCoordinates(height=0.4, width=0.4, top_left_x=0.0, top_left_y=0.0),
                "cherry_type": "King",
                "king_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "cherry_003",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.5, top_left_y=0.0),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_004",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.7, top_left_y=0.0),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Morello",
            },
        ],
    },
    "cherries-ig": {
        0: {
            "label_ref": "cherry_005",
            "coordinates": BoundingBoxCoordinates(height=0.4, width=0.4, top_left_x=0.0, top_left_y=0.0),
            "cherry_type": "Bing",
            "bing_quality_options": "Plump, Juicy",
        },
        2: [
            {
                "label_ref": "cherry_006",
                "coordinates": BoundingBoxCoordinates(height=0.1, width=0.2, top_left_x=0.5, top_left_y=0.0),
                "cherry_type": "King",
                "king_quality_options": "Large",
            },
            {
                "label_ref": "cherry_007",
                "coordinates": BoundingBoxCoordinates(height=0.1, width=0.2, top_left_x=0.0, top_left_y=0.5),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_008",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.8, top_left_y=0.0),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Queen Anne",
            },
        ],
    },
    "cherries-is": {
        0: {
            "label_ref": "cherry_009",
            "coordinates": BoundingBoxCoordinates(height=0.4, width=0.4, top_left_x=0.0, top_left_y=0.0),
            "cherry_type": "Bing",
            "bing_quality_options": "Plump",
        },
        3: [
            {
                "label_ref": "cherry_010",
                "coordinates": BoundingBoxCoordinates(height=0.4, width=0.4, top_left_x=0.5, top_left_y=0.0),
                "cherry_type": "King",
                "king_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "cherry_011",
                "coordinates": BoundingBoxCoordinates(height=0.3, width=0.3, top_left_x=0.0, top_left_y=0.5),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_012",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.9, top_left_y=0.0),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Lambert",
            },
        ],
    },
    "cherries-vid-001.mp4": {
        103: [
            {
                "label_ref": "cherry_013",
                "coordinates": BoundingBoxCoordinates(height=0.5, width=0.5, top_left_x=0.0, top_left_y=0.0),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_014",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.2, top_left_x=0.6, top_left_y=0.0),
                "cherry_type": "Bing",
                "bing_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "cherry_015",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.8, top_left_y=0.6),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Sweetheart",
            },
        ],
        104: [
            {
                "label_ref": "cherry_016",
                "coordinates": BoundingBoxCoordinates(height=0.5, width=0.5, top_left_x=0.0, top_left_y=0.0),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_014",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.2, top_left_x=0.6, top_left_y=0.0),
                "cherry_type": "Bing",
                "bing_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "cherry_017",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.8, top_left_y=0.6),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Sweetheart",
            },
        ],
        105: [
            {
                "label_ref": "cherry_016",
                "coordinates": BoundingBoxCoordinates(height=0.5, width=0.5, top_left_x=0.0, top_left_y=0.0),
                "cherry_type": "Rainier",
                "rainier_quality_options": "Plump",
            },
            {
                "label_ref": "cherry_014",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.2, top_left_x=0.6, top_left_y=0.0),
                "cherry_type": "Bing",
                "bing_quality_options": "Plump, Juicy, Large",
            },
            {
                "label_ref": "cherry_017",
                "coordinates": BoundingBoxCoordinates(height=0.2, width=0.1, top_left_x=0.8, top_left_y=0.6),
                "cherry_type": "Other cherry type",
                "Specify cherry type": "Sweetheart",
            },
        ],
    },
}

# Cache initialized label rows
label_row_map = {}

# Step 1: Initialize all label rows using a bundle
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for data_unit in video_frame_labels.keys():
        label_rows = project.list_label_rows_v2(data_title_eq=data_unit)
        assert isinstance(label_rows, list), f"Expected list of label rows for '{data_unit}', got {type(label_rows)}"

        if not label_rows:
            print(f"Skipping: No label row found for {data_unit}")
            continue

        label_row = label_rows[0]
        label_row.initialise_labels(bundle=bundle)
        assert label_row.ontology_structure is not None, f"Ontology not initialized for label row: {data_unit}"

        label_row_map[data_unit] = label_row  # Cache initialized label row for later use

# Step 2: Process all frames/annotations and prepare label rows to save
label_rows_to_save = []

for data_unit, frame_coordinates in video_frame_labels.items():
    label_row = label_row_map.get(data_unit)
    if not label_row:
        print(f"Skipping: No initialized label row found for {data_unit}")
        continue

    object_instances_by_label_ref = {}

    # Loop through the frames for the current data unit
    for frame_number, items in frame_coordinates.items():
        assert isinstance(frame_number, int), f"Frame number must be int, got {type(frame_number)}"
        if not isinstance(items, list):
            items = [items]

        for item in items:
            label_ref = item["label_ref"]
            coord = item["coordinates"]
            cherry_type = item["cherry_type"]

            assert cherry_type in {
                "Bing",
                "King",
                "Rainier",
                "Other cherry type",
            }, f"Unexpected cherry type '{cherry_type}' in data unit '{data_unit}'"

            # Check if label_ref already exists for reusability
            if label_ref not in object_instances_by_label_ref:
                box_object_instance: ObjectInstance = box_ontology_object.create_instance()
                assert box_object_instance is not None, "Failed to create ObjectInstance"

                object_instances_by_label_ref[label_ref] = box_object_instance
                checklist_attribute = None

                # Set cherry type attribute
                if cherry_type == "Bing":
                    assert bing_option is not None, "Missing 'bing_option'"
                    box_object_instance.set_answer(attribute=cherry_type_radio_attribute, answer=bing_option)
                    checklist_attribute = bing_checklist_attribute
                elif cherry_type == "King":
                    assert king_option is not None, "Missing 'king_option'"
                    box_object_instance.set_answer(attribute=cherry_type_radio_attribute, answer=king_option)
                    checklist_attribute = king_checklist_attribute
                elif cherry_type == "Rainier":
                    assert rainier_option is not None, "Missing 'rainier_option'"
                    box_object_instance.set_answer(attribute=cherry_type_radio_attribute, answer=rainier_option)
                    checklist_attribute = rainier_checklist_attribute
                elif cherry_type == "Other cherry type":
                    assert other_cherry_option is not None, "Missing 'other_cherry_option'"
                    box_object_instance.set_answer(attribute=cherry_type_radio_attribute, answer=other_cherry_option)
                    text_answer = item.get("Specify cherry type", "")
                    assert isinstance(text_answer, str), "'Specify cherry type' must be a string"
                    box_object_instance.set_answer(attribute=other_cherry_option_text_attribute, answer=text_answer)

                # Set checklist attributes
                checklist_answers = []
                quality_key = f"{cherry_type.lower()}_quality_options"
                quality_options = item.get(quality_key, "").split(", ")

                for quality in quality_options:
                    if quality == "Plump":
                        checklist_answers.append(
                            bing_plump_option
                            if cherry_type == "Bing"
                            else king_plump_option
                            if cherry_type == "King"
                            else rainier_plump_option
                        )
                    elif quality == "Juicy":
                        checklist_answers.append(
                            bing_juicy_option
                            if cherry_type == "Bing"
                            else king_juicy_option
                            if cherry_type == "King"
                            else rainier_juicy_option
                        )
                    elif quality == "Large":
                        checklist_answers.append(
                            bing_large_option
                            if cherry_type == "Bing"
                            else king_large_option
                            if cherry_type == "King"
                            else rainier_large_option
                        )

                if checklist_attribute and checklist_answers:
                    box_object_instance.set_answer(
                        attribute=checklist_attribute, answer=checklist_answers, overwrite=True
                    )

            else:
                # Reuse existing instance across frames
                box_object_instance = object_instances_by_label_ref[label_ref]

            # Assign the object to the frame and track it
            box_object_instance.set_for_frames(coordinates=coord, frames=frame_number)

    # Add object instances to label_row **only if they have frames assigned**
    for box_object_instance in object_instances_by_label_ref.values():
        assert isinstance(box_object_instance, ObjectInstance), "Expected ObjectInstance type"
        if box_object_instance.get_annotation_frames():  # Ensures it has at least one frame
            label_row.add_object_instance(box_object_instance)

    label_rows_to_save.append(label_row)

# Step 3: Save all label rows using a bundle
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows_to_save:
        assert label_row is not None, "Trying to save a None label row"
        label_row.save(bundle=bundle)
        print(f"Saved label row for {label_row.data_title}")

print("Labels with cherry type radio buttons, checklist attributes, and text labels added for all data units.")



# Label row constants

LABELS = "labels"
OBJECTS = "objects"
CLASSIFICATIONS = "classifications"
OBJECT_HASH = "objectHash"
CLASSIFICATION_HASH = "classificationHash"
OBJECT_ANSWERS = "object_answers"
CLASSIFICATION_ANSWERS = "classification_answers"


# Labeling algorithm names
INTERPOLATION = "interpolation"


# Type of Cord API key
TYPE_PROJECT = "project"
TYPE_DATASET = "dataset"
TYPE_ONTOLOGY = "ontology"



from dataclasses import dataclass
from typing import Optional

from encord.exceptions import ExceptionContext

HEADER_USER_AGENT = "User-Agent"
HEADER_CLOUD_TRACE_CONTEXT = "X-Cloud-Trace-Context"


@dataclass
class RequestContext(ExceptionContext):
    trace_id: Optional[str] = None
    span_id: Optional[str] = None
    domain: Optional[str] = None



"""
Code Block Name: All attributes
"""

# Import dependencies
import json
from collections.abc import Iterable

from encord import EncordUserClient
from encord.objects import ObjectInstance
from encord.objects.attributes import Attribute, ChecklistAttribute, RadioAttribute, TextAttribute

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
# SSH_PATH = get_ssh_key() # replace it with SSH key
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
DATA_UNIT = "cherries-010.jpg"
OUTPUT_FILE_PATH = "/Users/chris-encord/all_attributes_output.json"
BUNDLE_SIZE = 100

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Load project and label rows
project = user_client.get_project(PROJECT_ID)
assert project is not None, f"Project with ID {PROJECT_ID} could not be loaded"

label_rows = project.list_label_rows_v2(data_title_eq=DATA_UNIT)
assert label_rows, f"No label rows found for data unit: {DATA_UNIT}"

# Initialize label rows using bundle
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows:
        label_row.initialise_labels(bundle=bundle)

# Container for structured results
results = []


# Function to collect and print attributes
def extract_and_print_attributes(attribute: Attribute, object_instance: ObjectInstance, frame_number: int):
    attr_data = {
        "frame": frame_number + 1,
        "attribute_name": attribute.title,
        "attribute_hash": attribute.feature_node_hash,
        "attribute_type": None,
        "answers": [],
    }

    if isinstance(attribute, TextAttribute):
        attr_data["attribute_type"] = "TextAttribute"
        text_answer = object_instance.get_answer(attribute)
        attr_data["answers"].append({"text": text_answer or "No attribute answer"})

    elif isinstance(attribute, RadioAttribute):
        attr_data["attribute_type"] = "RadioAttribute"
        answer = object_instance.get_answer(attribute)
        if answer:
            answer_data = {"title": answer.title, "hash": answer.feature_node_hash, "nested_attributes": []}

            if hasattr(answer, "attributes") and answer.attributes:
                for nested_attribute in answer.attributes:
                    nested_result = extract_and_print_attributes(nested_attribute, object_instance, frame_number)
                    if nested_result:
                        answer_data["nested_attributes"].append(nested_result)

            attr_data["answers"].append(answer_data)
        else:
            attr_data["answers"].append({"note": "No attribute answer"})

    elif isinstance(attribute, ChecklistAttribute):
        attr_data["attribute_type"] = "ChecklistAttribute"
        answers = object_instance.get_answer(attribute)
        if answers:
            if not isinstance(answers, Iterable):
                answers = [answers]
            for answer in answers:
                attr_data["answers"].append({"title": answer.title, "hash": answer.feature_node_hash})
        else:
            attr_data["answers"].append({"note": "No attribute answer"})

    # Print the same info (optional)
    print(f"Frame {attr_data['frame']}: {attr_data['attribute_type']} name: {attr_data['attribute_name']}")
    for answer in attr_data["answers"]:
        if "title" in answer:
            print(f"Answer: {answer['title']}")
            print(f"Answer hash: {answer['hash']}")
        elif "text" in answer:
            print(f"Answer: {answer['text']}")
        elif "note" in answer:
            print(f"Note: {answer['note']}")
        if "nested_attributes" in answer:
            for nested in answer["nested_attributes"]:
                print(f"Nested attribute: {nested['attribute_name']} ({nested['attribute_hash']})")

    return attr_data


# Process all label rows
for label_row in label_rows:
    object_instances = label_row.get_object_instances()
    assert object_instances, f"No object instances found in label row {label_row.uid}"

    for object_instance in object_instances:
        annotations = object_instance.get_annotations()
        assert annotations, f"No annotations found for object instance {object_instance.object_hash}"

        assert object_instance.ontology_item and object_instance.ontology_item.attributes, (
            f"No attributes found for object {object_instance.object_hash}"
        )

        for annotation in annotations:
            for attribute in object_instance.ontology_item.attributes:
                extracted = extract_and_print_attributes(attribute, object_instance, annotation.frame)
                if extracted:
                    results.append(extracted)

# Save to JSON
assert OUTPUT_FILE_PATH.endswith(".json"), "Output file path must be a JSON file"

with open(OUTPUT_FILE_PATH, "w") as f:
    json.dump(results, f, indent=4)

print(f"\nAll attribute data saved to: {OUTPUT_FILE_PATH}")



import uuid
from unittest.mock import MagicMock, PropertyMock, patch

from encord.client import EncordClientProject
from encord.http.v2.api_client import ApiClient
from encord.http.v2.payloads import Page
from encord.orm.label_row import LabelRow
from encord.orm.project import Project as OrmProject
from encord.orm.project import ProjectDataset
from encord.project import Project
from tests.fixtures import ontology, project, user_client

assert user_client and project and ontology


UID = "d958ddbb-fcd0-477a-adf9-de14431dbbd2"


@patch.object(EncordClientProject, "get_project")
def test_label_rows_property_queries_metadata(project_client_mock: MagicMock, project: Project):
    project_current_orm_mock = MagicMock(spec=OrmProject)
    type(project_current_orm_mock).label_rows = PropertyMock(return_value=None)
    project._project_instance = project_current_orm_mock

    project_orm_mock = MagicMock(spec=OrmProject)
    project_client_mock.return_value = project_orm_mock
    type(project_orm_mock).label_rows = PropertyMock(return_value=[LabelRow({"data_title": "abc"})])

    project_client_mock.assert_not_called()

    rows = project.label_rows

    # Expect project data query to happen during the property call
    project_client_mock.assert_called_once()

    assert project_client_mock.call_args[1] == {"include_labels_metadata": True}

    assert len(rows) == 1
    assert rows[0].data_title == "abc"

    # Expect label rows metadata to be cached, so data query doesn't happen again
    _ = project.label_rows


@patch.object(ApiClient, "get")
def test_project_datasets(api_get: MagicMock, project: Project) -> None:
    dataset_hash = uuid.uuid4()
    expected_dataset = ProjectDataset(dataset_hash=dataset_hash, title="test dataset", description="my test dataset")
    api_get.return_value = Page(results=[expected_dataset])

    assert len(list(project.list_datasets())) == 1
    assert list(project.list_datasets()) == [expected_dataset]

    # Correctly serialised for legacy endpoint
    assert len(project.datasets) == 1
    assert project.datasets[0] == {
        "dataset_hash": str(dataset_hash),
        "title": "test dataset",
        "description": "my test dataset",
    }



"""---
title: "Dataset"
slug: "sdk-ref-dataset"
hidden: false
metadata:
  title: "Dataset"
  description: "Encord SDK Dataset class"
category: "64e481b57b6027003f20aaa0"
---
"""

from datetime import datetime
from pathlib import Path
from typing import Collection, Dict, Iterable, List, Optional, TextIO, Union
from uuid import UUID

from encord.client import EncordClientDataset
from encord.common.deprecated import deprecated
from encord.constants.enums import DataType
from encord.http.utils import CloudUploadSettings
from encord.orm.cloud_integration import CloudIntegration
from encord.orm.dataset import (
    AddPrivateDataResponse,
    DataLinkDuplicatesBehavior,
    DataRow,
    DatasetAccessSettings,
    DatasetDataLongPolling,
    DatasetUser,
    DatasetUserRole,
    Image,
    ImageGroup,
    ImageGroupOCR,
    StorageLocation,
    Video,
)
from encord.orm.dataset import Dataset as OrmDataset
from encord.orm.group import DatasetGroup
from encord.storage import StorageFolder
from encord.utilities.hash_utilities import convert_to_uuid


class Dataset:
    """Access dataset-related data and manipulate the dataset."""

    def __init__(self, client: EncordClientDataset, orm_dataset: OrmDataset):
        self._client = client
        self._dataset_instance = orm_dataset

    @property
    def dataset_hash(self) -> str:
        """Get the dataset hash (i.e. the Dataset ID).

        Returns:
            str: The dataset hash.
        """
        return self._dataset_instance.dataset_hash

    @property
    def title(self) -> str:
        """Get the title of the dataset.

        Returns:
            str: The title of the dataset.
        """
        return self._dataset_instance.title

    @property
    def description(self) -> str:
        """Get the description of the dataset.

        Returns:
            str: The description of the dataset.
        """
        return self._dataset_instance.description

    @property
    def storage_location(self) -> StorageLocation:
        """Get the storage location of the dataset.

        Returns:
            StorageLocation: The storage location of the dataset.
        """
        return self._dataset_instance.storage_location

    @property
    def backing_folder_uuid(self) -> Optional[UUID]:
        """Get the UUID of the backing folder.

        Returns:
            Optional[UUID]: The UUID of the backing folder, if any.
        """
        return self._dataset_instance.backing_folder_uuid

    @property
    def data_rows(self) -> List[DataRow]:
        """Get the data rows of the dataset.

        Part of the response of this function can be configured by the
        :meth:`encord.dataset.Dataset.set_access_settings` method.

        .. code::

            dataset.set_access_settings(DatasetAccessSettings(fetch_client_metadata=True))
            print(dataset.data_rows)

        Returns:
            List[DataRow]: A list of DataRow objects.
        """
        return self._dataset_instance.data_rows

    def list_data_rows(
        self,
        title_eq: Optional[str] = None,
        title_like: Optional[str] = None,
        created_before: Optional[Union[str, datetime]] = None,
        created_after: Optional[Union[str, datetime]] = None,
        data_types: Optional[List[DataType]] = None,
        data_hashes: Optional[List[str]] = None,
    ) -> List[DataRow]:
        """Retrieve dataset rows (pointers to data, labels).

        Args:
            title_eq: Optional exact title row filter.
            title_like: Optional fuzzy title row filter; SQL syntax.
            created_before: Optional datetime row filter.
            created_after: Optional datetime row filter.
            data_types: Optional data types row filter.
            data_hashes: Optional list of individual data unit hashes to include.

        Returns:
            List[DataRow]: A list of DataRow objects that match the filter.

        Raises:
            AuthorisationError: If the dataset API key is invalid.
            ResourceNotFoundError: If no dataset exists by the specified dataset EntityId.
            UnknownError: If an error occurs while retrieving the dataset.
        """
        return self._client.list_data_rows(title_eq, title_like, created_before, created_after, data_types, data_hashes)

    def refetch_data(self) -> None:
        """Refetch the dataset properties.

        The Dataset class will only fetch its properties once. Use this function if you suspect the state of those
        properties to be outdated.
        """
        self._dataset_instance = self._client.get_dataset()

    def get_dataset(self) -> OrmDataset:
        """Get the dataset instance.

        This function is exposed for convenience. You are encouraged to use the property accessors instead.

        Returns:
            OrmDataset: The dataset instance.
        """
        return self._client.get_dataset()

    def set_access_settings(self, dataset_access_settings: DatasetAccessSettings, *, refetch_data: bool = True) -> None:
        """Set access settings for the dataset.

        Args:
            dataset_access_settings: The access settings to use going forward.
            refetch_data: Whether a `refetch_data()` call should follow the update of the dataset access settings.
        """
        self._client.set_access_settings(dataset_access_settings)
        if refetch_data:
            self.refetch_data()

    def add_users(self, user_emails: List[str], user_role: DatasetUserRole) -> List[DatasetUser]:
        """Add users to the dataset.


        If the user already exists in the Dataset, this operation succeeds but the `user_role` remains unchanged. The
        existing `user_role` is reflected in the `DatasetUser` instance returned.

        Args:
            user_emails: List of user emails to be added.
            user_role: The user role to assign to all users.

        Returns:
            List[DatasetUser]: A list of DatasetUser instances reflecting the added users.
        """
        return self._client.add_users(user_emails, user_role)

    def list_groups(self) -> Iterable[DatasetGroup]:
        """List all groups that have access to the dataset.

        Returns:
            Iterable[DatasetGroup]: An iterable of DatasetGroup instances.
        """
        dataset_hash = convert_to_uuid(self.dataset_hash)
        page = self._client.list_groups(dataset_hash)
        yield from page.results

    def add_group(self, group_hash: Union[List[UUID], UUID], user_role: DatasetUserRole) -> None:
        """Add a group to the dataset.

        Args:
            group_hash: List of group hashes to be added.
            user_role: The user role that the group will be given.

        Returns:
            None
        """
        if isinstance(group_hash, UUID):
            group_hash = [group_hash]
        self._client.add_groups(self.dataset_hash, group_hash, user_role)

    def remove_group(self, group_hash: Union[List[UUID], UUID]) -> None:
        """Remove a group from the dataset.

        Args:
            group_hash: List of group hashes to be removed.

        Returns:
            None
        """
        if isinstance(group_hash, UUID):
            group_hash = [group_hash]
        dataset_hash = convert_to_uuid(self.dataset_hash)
        self._client.remove_groups(dataset_hash, group_hash)

    def upload_video(
        self,
        file_path: Union[str, Path],
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        folder: Optional[Union[UUID, StorageFolder]] = None,
    ) -> Video:
        """Upload a video to Encord storage.

        Args:
            file_path: Path to the video, e.g., '/home/user/data/video.mp4'.
            cloud_upload_settings: Settings for uploading data into the cloud. Change this object to overwrite the default values.
            title: The video title. If unspecified, this will be the file name. This title should include an extension. For example: "encord_video.mp4".
            folder: When uploading to a non-mirror dataset, you have to specify the folder to store the file in. This can be either a :class:`encord.storage.Folder` instance or the UUID of the folder.

        Returns:
            Video: An object describing the created video, see :class:`encord.orm.dataset.Video`.

        Raises:
            UploadOperationNotSupportedError: If trying to upload to external datasets (e.g., S3/GPC/Azure).
        """
        folder_uuid = folder.uuid if isinstance(folder, StorageFolder) else folder

        return self._client.upload_video(
            file_path, cloud_upload_settings=cloud_upload_settings, title=title, folder_uuid=folder_uuid
        )

    def create_image_group(
        self,
        file_paths: Iterable[Union[str, Path]],
        max_workers: Optional[int] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        *,
        create_video: bool = True,
        folder: Optional[Union[UUID, StorageFolder]] = None,
    ) -> List[ImageGroup]:
        """Create an image group in Encord storage.

        Choose this type of image upload for sequential images. Else, you can choose the :meth:`.Dataset.upload_image` function.

        Args:
            file_paths: A list of paths to images, e.g., ['/home/user/data/img1.png', '/home/user/data/img2.png'].
            max_workers: DEPRECATED: This argument will be ignored.
            cloud_upload_settings: Settings for uploading data into the cloud. Change this object to overwrite the default values.
            title: The title of the image group. If unspecified, this will be randomly generated for you. This title should NOT include an extension. For example, "encord_image_group".
            create_video: A flag specifying how image groups are stored. If `True`, a compressed video will be created from the image groups. `True` was the previous default support. If `False`, the images are saved as a sequence of images.
            folder: When uploading to a non-mirror dataset, you have to specify the folder to store the file in. This can be either a :class:`encord.storage.Folder` instance or the UUID of the folder.

        Returns:
            List[ImageGroup]: A list containing the object(s) describing the created data unit(s). See :class:`encord.orm.dataset.ImageGroup`.

        Raises:
            UploadOperationNotSupportedError: If trying to upload to external datasets (e.g., S3/GPC/Azure).
            InvalidArgumentError: If the folder is specified, but the dataset is a mirror dataset.
        """
        return self._client.create_image_group(
            file_paths,
            cloud_upload_settings=cloud_upload_settings,
            title=title,
            create_video=create_video,
            folder_uuid=folder.uuid if isinstance(folder, StorageFolder) else folder,
        )

    def create_dicom_series(
        self,
        file_paths: Union[Collection[str], Collection[Path], Collection[Union[Path, str]]],
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        title: Optional[str] = None,
        folder: Optional[Union[UUID, StorageFolder]] = None,
    ) -> Dict:
        """Upload a DICOM series to Encord storage.

        Args:
            file_paths: A collection of paths to DICOM files. Example: ['/home/user/data/DICOM_1.dcm', '/home/user/data/DICOM_2.dcm']
            cloud_upload_settings: Settings for uploading data into the cloud. Change this object to overwrite the default values.
            title: The title of the DICOM series. If unspecified, a random title will be generated. This title should NOT include an extension. Example: "encord_image_group"
            folder: When uploading to a non-mirror dataset, specify the folder to store the file in. This can be either a `StorageFolder` instance or the UUID of the folder.

        Returns:
            A dictionary describing the created series.

        Raises:
            UploadOperationNotSupportedError: If trying to upload to external datasets (e.g., S3/GPC/Azure).
            InvalidArgumentError: If the folder is specified, but the dataset is a mirror dataset.
        """
        return self._client.create_dicom_series(
            file_paths,
            cloud_upload_settings=cloud_upload_settings,
            title=title,
            folder_uuid=folder.uuid if isinstance(folder, StorageFolder) else folder,
        )

    def upload_image(
        self,
        file_path: Union[Path, str],
        title: Optional[str] = None,
        cloud_upload_settings: CloudUploadSettings = CloudUploadSettings(),
        folder: Optional[Union[UUID, StorageFolder]] = None,
    ) -> Image:
        """Upload a single image to Encord storage. For sequential images, consider creating an image group
        using the :meth:`.Dataset.create_image_group` function.

        Args:
            file_path: The file path to the image.
            title: The image title. If unspecified, the file name will be used. This title should include an extension. Example: "encord_image.png"
            cloud_upload_settings: Settings for uploading data into the cloud. Change this object to overwrite the default values.
            folder: When uploading to a non-mirror dataset, specify the folder to store the file in. This can be either a `StorageFolder` instance or the UUID of the folder.

        Returns:
            Uploaded Image object.

        """
        folder_uuid = folder.uuid if isinstance(folder, StorageFolder) else folder
        return self._client.upload_image(file_path, title, cloud_upload_settings, folder_uuid)

    def link_items(
        self,
        item_uuids: List[UUID],
        duplicates_behavior: DataLinkDuplicatesBehavior = DataLinkDuplicatesBehavior.SKIP,
    ) -> List[DataRow]:
        """Link storage items to the dataset, creating new data rows.

        Args:
            item_uuids: List of item UUIDs to link to the dataset.
            duplicates_behavior: The behavior to follow when encountering duplicates. Defaults to `SKIP`. See also :class:`encord.orm.dataset.DataLinkDuplicatesBehavior`.

        Returns:
            List of DataRow objects representing linked items.

        """
        return self._client.link_items(item_uuids, duplicates_behavior)

    def delete_data(self, data_hashes: Union[List[str], str]):
        """Delete a video/image group from a dataset.

        Args:
            data_hashes: List of hashes of the videos/image_groups you'd like to delete. All should belong to the same dataset.

        """
        return self._client.delete_data(data_hashes)

    def add_private_data_to_dataset(
        self,
        integration_id: str,
        private_files: Union[str, Dict, Path, TextIO],
        ignore_errors: bool = False,
    ) -> AddPrivateDataResponse:
        """Append data hosted on a private cloud to an existing dataset.

        For a more complete example of safe uploads, please follow the guide found in our docs under
        :ref:`https://python.docs.encord.com/tutorials/datasets.html#adding-data-from-a-private-cloud`

        Args:
            integration_id: The `EntityId` of the cloud integration you wish to use.
            private_files: A path to a JSON file, JSON string, Python dictionary, or a `Path` object containing the files you wish to add.
            ignore_errors: When set to `True`, prevent individual errors from stopping the upload process.

        Returns:
            List of DatasetDataInfo objects containing data_hash and title.

        """
        return self._client.add_private_data_to_dataset(integration_id, private_files, ignore_errors)

    def add_private_data_to_dataset_start(
        self,
        integration_id: str,
        private_files: Union[str, Dict, Path, TextIO],
        ignore_errors: bool = False,
        *,
        folder: Optional[Union[StorageFolder, UUID]] = None,
    ) -> str:
        """Append data hosted on a private cloud to an existing dataset.

        This method initializes the upload in Encord's backend.
        Once the upload ID has been returned, you can exit the terminal
        while the job continues uninterrupted.

        You can check upload job status at any point using
        the :meth:`add_private_data_to_dataset_get_result` method.
        This can be done in a separate Python session to the one
        where the upload was initialized.

        Args:
            integration_id: The `EntityId` of the cloud integration you wish to use.
            private_files:A path to a JSON file, JSON string, Python dictionary, or a `Path` object containing the files you wish to add.
            ignore_errors: When set to `True`, prevent individual errors from stopping the upload process.
            folder: When uploading to a non-mirror dataset, specify the folder to store the file in. This can be either a `StorageFolder` instance or the UUID of the folder.

        Returns:
            UUID Identifier of the upload job.
            This ID enables the user to track the job progress via SDK or web app.

        """
        folder_uuid = folder.uuid if isinstance(folder, StorageFolder) else folder
        return self._client.add_private_data_to_dataset_start(integration_id, private_files, ignore_errors, folder_uuid)

    def add_private_data_to_dataset_get_result(
        self,
        upload_job_id: str,
        timeout_seconds: int = 7 * 24 * 60 * 60,  # 7 days
    ) -> DatasetDataLongPolling:
        """Fetch data upload status, perform long polling process for `timeout_seconds`.

        Args:
            upload_job_id: UUID Identifier of the upload job. This ID enables the user to track the job progress via SDK or web app.
            timeout_seconds: Number of seconds the method will wait while waiting for a response. If `timeout_seconds == 0`, only a single checking request is performed. Response will be immediately returned.

        Returns:
            Response containing details about job status, errors, and progress.

        """
        return self._client.add_private_data_to_dataset_get_result(upload_job_id, timeout_seconds)

    def update_data_item(self, data_hash: str, new_title: str) -> bool:
        """DEPRECATED: Use the individual setter properties of the respective :class:`encord.orm.dataset.DataRow`
        instance instead. These can be retrieved via the :meth:`.Dataset.data_rows` function.

        Update a data item.

        Args:
            data_hash: Data hash of the item being updated.
            new_title: New title of the data item being updated.

        Returns:
        Boolean indicating whether the update was successful.

        """
        return self._client.update_data_item(data_hash, new_title)

    def re_encode_data(self, data_hashes: List[str]):
        """Launch an async task that can re-encode a list of videos.

        Args:
            data_hashes: List of hashes of the videos you'd like to re-encode. All should belong to the same dataset.

        Returns:
            Entity ID of the async task launched.

        """
        return self._client.re_encode_data(data_hashes)

    def re_encode_data_status(self, job_id: int):
        """Returns the status of an existing async task aimed at re-encoding videos.

        Args:
            job_id: ID of the async task that was launched to re-encode the videos.

        Returns:
            Object containing the status of the task, along with info about the new encoded videos
            in case the task has been completed.

        """
        return self._client.re_encode_data_status(job_id)

    def run_ocr(self, image_group_id: str) -> List[ImageGroupOCR]:
        """Returns an optical character recognition result for a given image group.

        Args:
            image_group_id: The ID of the image group in this dataset to run OCR on.

        Returns:
            List of ImageGroupOCR objects representing the text and corresponding coordinates
            found in each frame of the image group.

        """
        return self._client.run_ocr(image_group_id)

    @deprecated(version="0.1.154", alternative="EncordUserClient.get_cloud_integrations")
    def get_cloud_integrations(self) -> List[CloudIntegration]:
        """Retrieve a list of cloud integrations configured in Encord.

        Returns:
            List of CloudIntegration objects representing configured cloud integrations.

        """
        return self._client.get_cloud_integrations()



"""
Code Block Name: Create Data Groups
"""

from uuid import UUID

from encord.constants.enums import DataType
from encord.objects.metadata import DataGroupMetadata
from encord.orm.storage import DataGroupCustom, StorageItemType
from encord.user_client import EncordUserClient

# --- Configuration ---
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"  # Replace with the file path to your SSH key
FOLDER_ID = "00000000-0000-0000-0000-000000000000"  # Replace with the Folder ID
PROJECT_ID = "00000000-0000-0000-0000-000000000000"  # Replace with the Project ID
DATASET_ID = "00000000-0000-0000-0000-000000000000"  # Replace with the Dataset ID


# --- Connect to Encord ---
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

folder = user_client.get_storage_folder(FOLDER_ID)

# --- Reusable layout and settings ---
layout = {
    "direction": "column",
    "first": {"type": "data_unit", "key": "instructions"},
    "second": {
        "direction": "column",
        "first": {
            "direction": "row",
            "first": {"type": "data_unit", "key": "top-left"},
            "second": {"type": "data_unit", "key": "top-right"},
            "splitPercentage": 50,
        },
        "second": {
            "direction": "row",
            "first": {"type": "data_unit", "key": "bottom-left"},
            "second": {"type": "data_unit", "key": "bottom-right"},
            "splitPercentage": 50,
        },
        "splitPercentage": 50,
    },
    "splitPercentage": 20,
}
settings = {"tile_settings": {"instructions": {"is_read_only": True}}}

# --- List of data groups with specific UUIDs per tile ---
groups = [
    {
        "name": "group-01",
        "uuids": {
            "instructions": UUID("4b5fb04c-4626-4a6b-86bc-09ea6a6893f8"),  # Data unit ID for text file or PDF
            "top-left": UUID("0d36238a-2759-42c3-8e4e-a788f9425ee4"),  # Data unit ID for image, video, or audio file
            "top-right": UUID("62fff5b4-df82-43fe-84b7-1bcfc090c14e"),  # Data unit ID for image, video, or audio file
            "bottom-left": UUID("e69352c6-e1db-4b4a-a9d7-0fde9fc84cf1"),  # Data unit ID for image, video, or audio file
            "bottom-right": UUID(
                "a8b61faf-f5dd-4543-ac15-f1e95328c15c"
            ),  # Data unit ID for image, video, or audio file
        },
    },
    {
        "name": "group-02",
        "uuids": {
            "instructions": UUID("209ecfc2-73fc-414c-8627-3ad7eb542e49"),  # Data unit ID for text file or PDF
            "top-left": UUID("2a88145a-e8a5-4c68-a9f1-c34b32f24d62"),  # Data unit ID for image, video, or audio file
            "top-right": UUID("88ee8731-25de-4571-8822-78d4ecfc2dde"),  # Data unit ID for image, video, or audio file
            "bottom-left": UUID("7ebfbd65-965a-49f7-926b-108b594f8d63"),  # Data unit ID for image, video, or audio file
            "bottom-right": UUID(
                "20983fbf-666e-4626-b1e3-fc3c206313c1"
            ),  # Data unit ID for image, video, or audio file
        },
    },
    # More groups...
]

for g in groups:
    group = folder.create_data_group(
        DataGroupCustom(
            name=g["name"],
            layout=layout,
            layout_contents=g["uuids"],
            settings=settings,
        )
    )
    print(f"✅ Created group '{g['name']}' with UUID {group}")

# Add all the data groups in a folder to a dataset
group_items = folder.list_items(item_types=[StorageItemType.GROUP])
d = user_client.get_dataset(DATASET_ID)
d.link_items([item.uuid for item in group_items])

p = user_client.get_project(PROJECT_ID)
rows = p.list_label_rows_v2(include_children=True)

# label rows of data groups have a "metadata" attribute
for row in rows:
    if row.data_type == DataType.GROUP:
        row.initialise_labels()
        assert isinstance(row.metadata, DataGroupMetadata)
        print(row.metadata.children)



from unittest.mock import MagicMock, patch

import pytest
from requests import Response, Session

from encord.configs import SshConfig
from encord.exceptions import EncordException
from encord.http.v2.api_client import ApiClient
from encord.orm.analytics import CollaboratorTimer
from tests.fixtures import PRIVATE_KEY


@pytest.fixture
def api_client():
    return ApiClient(config=SshConfig(PRIVATE_KEY))


@patch.object(Session, "send")
def test_deserialise_payload_raises_on_wrong_payload(send: MagicMock, api_client: ApiClient):
    res = Response()
    res.status_code = 200
    res._content = b'{ "some_wrong_key": "some_wrong_value" }'
    send.return_value = res

    with pytest.raises(EncordException):
        api_client.get("/", params=None, result_type=CollaboratorTimer)


@patch.object(Session, "send")
def test_deserialise_payload_ok_on_correct_payload(send: MagicMock, api_client: ApiClient):
    res = Response()
    res.status_code = 200
    res._content = b"""
    {
        "user_email": "noone@nowhere.com",
        "user_role": 0,
        "data_title": "some title",
        "time_seconds": 123.456
    }
    """
    send.return_value = res

    res = api_client.get("/", params=None, result_type=CollaboratorTimer)

    assert isinstance(res, CollaboratorTimer)
    assert res == CollaboratorTimer(
        user_email="noone@nowhere.com", user_role=0, data_title="some title", time_seconds=123.456
    )


@patch.object(Session, "send")
def test_deserialise_payload_ok_on_extra_keys(send: MagicMock, api_client: ApiClient):
    res = Response()
    res.status_code = 200
    res._content = b"""
    {
        "user_email": "noone@nowhere.com",
        "user_role": 0,
        "data_title": "some title",
        "time_seconds": 123.456,
        "extra_key": "extra_value"
    }
    """
    send.return_value = res

    res = api_client.get("/", params=None, result_type=CollaboratorTimer)

    assert isinstance(res, CollaboratorTimer)
    assert res == CollaboratorTimer(
        user_email="noone@nowhere.com", user_role=0, data_title="some title", time_seconds=123.456
    )



"""---
title: "Exceptions"
slug: "sdk-ref-exceptions"
hidden: false
metadata:
  title: "Exceptions"
  description: "Encord SDK Exceptions."
category: "64e481b57b6027003f20aaa0"
---
"""

from dataclasses import dataclass, field, fields
from datetime import datetime, timezone
from typing import Optional


@dataclass
class ExceptionContext:
    timestamp: str = field(default_factory=lambda: datetime.now(tz=timezone.utc).isoformat())

    def __str__(self):
        return " ".join(
            f"{object_field.name}={getattr(self, object_field.name)!r}"
            for object_field in fields(self)
            if getattr(self, object_field.name) is not None
        )


class EncordException(Exception):
    """Base class for all exceptions."""

    def __init__(self, message: str, context: Optional[ExceptionContext] = None):
        super().__init__(message)
        self.message = message
        self.context = context if context is not None else ExceptionContext()

    def __str__(self):
        return f"{self.message} {self.context}"


CordException = EncordException


class InitialisationError(EncordException):
    """Exception thrown when API key fails to initialise."""

    pass


class AuthenticationError(EncordException):
    """Exception thrown when API key fails authentication."""

    pass


class AuthorisationError(EncordException):
    """Exception thrown when access is unauthorised.
    (E.g. access to a data asset or method).
    """

    pass


class ResourceNotFoundError(EncordException):
    """Exception thrown when a requested resource is not found.
    (E.g. label, data asset).
    """

    pass


class TimeOutError(EncordException):
    """Exception thrown when a request times out."""

    pass


class RequestException(EncordException):
    """Ambiguous exception while handling request."""

    pass


class RateLimitExceededError(EncordException):
    """Exception raised when requests exceed the Encord API rate limit."""

    def __init__(
        self,
        *,
        message: str = "Rate limit exceeded.",
        context: Optional[ExceptionContext] = None,
        retry_after: Optional[int] = None,
    ) -> None:
        if retry_after:
            message += f" Retry after {retry_after} second{'s' if retry_after > 1 else ''}."

        super().__init__(message, context)
        self.retry_after = retry_after


class UnknownException(EncordException):
    """Unknown error."""

    pass


class InvalidDateFormatError(EncordException):
    """Invalid date format error"""

    pass


class MethodNotAllowedError(EncordException):
    """Exception thrown when HTTP method is not allowed."""

    pass


class OperationNotAllowed(EncordException):
    """Exception thrown when a read/write operation is not allowed.
    The API key blocks the operation.
    """

    pass


class AnswerDictionaryError(EncordException):
    """Exception thrown when answer dictionaries are incomplete.
    Occurs when an object or classification is missing.
    """

    pass


class CorruptedLabelError(EncordException):
    """Exception thrown when a label is corrupted.
    (E.g. the frame labels have more frames than the video).
    """

    pass


class FileTypeNotSupportedError(EncordException):
    """Exception thrown when a file type is not supported.
    Supported file types are: image/jpeg, image/png, video/webm, video/mp4.
    """

    pass


class FileSizeNotSupportedError(EncordException):
    """Exception thrown when the combined size of the input files is larger than the supported limit."""

    pass


class FeatureDoesNotExistError(EncordException):
    """If a feature uid does not exist in a given project ontology."""

    pass


class ModelWeightsInconsistentError(EncordException):
    """Exception thrown when an attempted model training iteration has a different
    type of weights than what is recorded (i.e. if type of model_hash (uid) is faster_rcnn,
    but is attempted trained with different model weights).
    """

    pass


class ModelFeaturesInconsistentError(EncordException):
    """If a feature type is different than what is supported by the model (e.g. if
    creating a classification model using a bounding box).
    """

    pass


class UploadOperationNotSupportedError(EncordException):
    """Exception thrown when trying to upload a video/image group to non-Encord storage dataset"""

    pass


class DetectionRangeInvalidError(EncordException):
    """Exception thrown when a detection range is invalid.
    (E.g. negative or higher than num frames in video).
    """

    pass


class InvalidAlgorithmError(EncordException):
    """Exception thrown when invalid labeling algorithm name is sent."""

    pass


class ResourceExistsError(EncordException):
    """Exception thrown when trying to re-create a resource.
    Avoids overriding existing work.
    """

    pass


class DuplicateSshKeyError(EncordException):
    """Exception thrown when using an SSH key that was added twice to the platform."""

    pass


class SshKeyNotFound(EncordException):
    """Exception thrown when using an SSH key that was not added to the platform."""


class InvalidArgumentsError(EncordException):
    """Exception thrown when the arguments are invalid."""

    pass


class GenericServerError(EncordException):
    """The server has reported an error which is not recognised by this SDK version. Try upgrading the SDK version to
    see the precise error that is reported.
    """

    pass


class CloudUploadError(EncordException):
    """The upload to the cloud was not successful"""

    pass


class MultiLabelLimitError(EncordException):
    """Too many labels were requested"""

    def __init__(self, message, maximum_labels_allowed: int, context: Optional[ExceptionContext] = None):
        super().__init__(message=message, context=context)
        self.maximum_labels_allowed = maximum_labels_allowed


class LabelRowError(EncordException):
    """An error thrown when the construction of a LabelRow class is invalid."""

    pass


class OntologyError(EncordException):
    """An error thrown when using the ontology class with an error."""

    pass


class WrongProjectTypeError(CordException):
    """An error thrown when project type does not match the operation
    E.g. when TMS2 specific operations are attempted on non-TMS2 project
    """



"""
Code Block Name: Create Project
"""

# Import dependencies
from encord.user_client import EncordUserClient

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_TITLE = "My Project 01"
DATASET_ID_01 = "00000000-0000-0000-0000-000000000000"
DATASET_ID_02 = "00000000-0000-0000-0000-000000000000"
DATASET_ID_03 = "00000000-0000-0000-0000-000000000000"
WORKFLOW_TEMPLATE_ID = "00000000-0000-0000-0000-000000000000"

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Create a project
project = user_client.create_project(
    project_title=PROJECT_TITLE,
    dataset_hashes=[DATASET_ID_01, DATASET_ID_02, DATASET_ID_03],
    workflow_template_hash=WORKFLOW_TEMPLATE_ID,
)

# Prints the Project ID of the Project you just created
print(project)



"""
Code Block Name: View Project details
"""

# Import dependencies
from encord import EncordUserClient

# User input
SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Open the project you want to work on by specifying the Project ID
project = user_client.get_project(PROJECT_ID)

# Prints relevant Project information
print(f"Project Title: {project.title}")
print(f"Description: {project.description}")
print(f"Created at: {project.created_at}")
print(f"Ontology ID: {project.ontology_hash}")
print(f"Datasets: {project.list_datasets()}")
print(f"Workflow: {project.workflow.stages}")



"""
Code Block Name: Export labels to JSON
"""

# Import dependencies
import json
import os

from encord import EncordUserClient

SSH_PATH = "/Users/chris-encord/ssh-private-key.txt"
PROJECT_ID = "00000000-0000-0000-0000-000000000000"
BUNDLE_SIZE = 100  # Customize as needed

# Create user client using SSH key
user_client: EncordUserClient = EncordUserClient.create_with_ssh_private_key(
    ssh_private_key_path=SSH_PATH,
    # For US platform users use "https://api.us.encord.com"
    domain="https://api.encord.com",
)

# Specify Project
project = user_client.get_project(PROJECT_ID)
assert project is not None, f"Project with ID {PROJECT_ID} could not be loaded"

# Get label rows for your Project
label_rows = project.list_label_rows_v2()
assert label_rows, f"No label rows found in project {PROJECT_ID}"

# Initialize label rows using bundles
with project.create_bundle(bundle_size=BUNDLE_SIZE) as bundle:
    for label_row in label_rows:
        label_row.initialise_labels(bundle=bundle)

# Collect all label row data
all_label_rows = [label_row.to_encord_dict() for label_row in label_rows]
assert all_label_rows, "No label row data collected for export"

# Save the collected label rows data to a JSON file
output_file = "/Users/chris-encord/export-label-rows.json"
assert output_file.endswith(".json"), "Output file must be a .json file"

with open(output_file, "w") as file:
    json.dump(all_label_rows, file, indent=4)

print(f"Label rows have been saved to {output_file}.")



import copy
import logging

import pytest

import encord.objects
import encord.objects.attributes
import encord.objects.classification
import encord.objects.ontology_structure
import encord.objects.options
from encord.objects.common import Shape
from encord.objects.skeleton_template import SkeletonTemplate, SkeletonTemplateCoordinate
from encord.objects.utils import short_uuid_str
from tests.objects.data.data_editor_blob import EDITOR_BLOB

log = logging.getLogger(__name__)
log.setLevel(logging.INFO)

# intentionally using a different import for backwards compatibility check
OBJECT_1 = encord.objects.Object(
    uid=1,
    name="Eye",
    color="#D33115",
    shape=encord.objects.Shape.BOUNDING_BOX,
    feature_node_hash="a55abbeb",
)

FLAT_OPTION_1 = encord.objects.FlatOption(
    uid=[2, 1, 1],
    feature_node_hash="2bc17c88",
    label="Is it a cute nose?",
    value="is_it_a_cute_nose?",
)

FLAT_OPTION_2 = encord.objects.FlatOption(
    uid=[2, 1, 2],
    feature_node_hash="86eaa4f2",
    label="Is it a wet nose? ",
    value="is_it_a_wet_nose?_",
)

CHECKLIST_ATTRIBUTE = encord.objects.ChecklistAttribute(
    uid=[2, 1],
    feature_node_hash="1e3e5cad",
    name="Additional details about the nose",
    required=True,
    dynamic=False,
    options=[FLAT_OPTION_1, FLAT_OPTION_2],
)


OBJECT_2 = encord.objects.Object(
    uid=2,
    name="Nose",
    color="#E27300",
    shape=encord.objects.Shape.POLYGON,
    feature_node_hash="86648f32",
    attributes=[CHECKLIST_ATTRIBUTE],
)

# intentionally using a different import for backwards compatibility check
RADIO_ATTRIBUTE_1 = encord.objects.RadioAttribute(
    uid=[4, 1, 1, 1],
    feature_node_hash="59204845",
    name="Leaf",
    required=False,
    dynamic=False,
)

NESTABLE_OPTION_3 = encord.objects.NestableOption(
    uid=[4, 1, 1],
    feature_node_hash="5d102ce6",
    label="Nested Option",
    value="nested_option",
    nested_options=[RADIO_ATTRIBUTE_1],
)

RADIO_ATTRIBUTE_2 = encord.objects.RadioAttribute(
    uid=[4, 1],
    feature_node_hash="cabfedb5",
    name="Radio with options",
    required=False,
    dynamic=False,
    options=[NESTABLE_OPTION_3],
)
OBJECT_3 = encord.objects.Object(
    uid=3,
    name="Example",
    color="#FE9200",
    shape=encord.objects.Shape.POLYLINE,
    feature_node_hash="6eeba59b",
    attributes=[RADIO_ATTRIBUTE_2],
)

NESTABLE_OPTION_1 = encord.objects.NestableOption(
    uid=[1, 1, 1],
    feature_node_hash="a3aeb48d",
    label="Yes",
    value="yes",
)

NESTABLE_OPTION_2 = encord.objects.NestableOption(
    uid=[1, 1, 2],
    feature_node_hash="d0a4b373",
    label="No",
    value="no",
)

RADIO_ATTRIBUTE_3 = encord.objects.RadioAttribute(
    uid=[1, 1],
    feature_node_hash="a6136d14",
    name="Is the cat standing?",
    required=True,
    dynamic=False,
    options=[NESTABLE_OPTION_1, NESTABLE_OPTION_2],
)
# intentionally using a different import for backwards compatibility check
CLASSIFICATION_1 = encord.objects.Classification(
    uid=1,
    feature_node_hash="a39d81c0",
    attributes=[RADIO_ATTRIBUTE_3],
)
SKELETON_TEMPLATE_COORDINATES = [
    SkeletonTemplateCoordinate(x=0, y=0, name="point_0"),
    SkeletonTemplateCoordinate(x=1, y=1, name="point_1"),
]
SKELETON_TEMPLATE_LINE = SkeletonTemplate(
    name="Line",
    width=100,
    height=100,
    skeleton={str(i): x for (i, x) in enumerate(SKELETON_TEMPLATE_COORDINATES)},
    skeleton_edges={"0": {"1": {"color": "#00000"}}},
    feature_node_hash="c67522ee",
)
# intentionally using a different import for backwards compatibility check
EXPECTED_ONTOLOGY: encord.objects.OntologyStructure = encord.objects.OntologyStructure(
    objects=[
        OBJECT_1,
        OBJECT_2,
        OBJECT_3,
    ],
    classifications=[CLASSIFICATION_1],
    skeleton_templates={"Line": SKELETON_TEMPLATE_LINE},
)


def test_ontology_structure_round_trip():
    assert EXPECTED_ONTOLOGY == encord.objects.OntologyStructure.from_dict(EXPECTED_ONTOLOGY.to_dict())


def test_json_to_ontology():
    actual = encord.objects.OntologyStructure.from_dict(copy.deepcopy(EDITOR_BLOB))
    assert EXPECTED_ONTOLOGY == actual


def test_ontology_to_json():
    actual = EXPECTED_ONTOLOGY.to_dict()
    assert copy.deepcopy(EDITOR_BLOB) == actual


def test_add_classification():
    ontology = encord.objects.OntologyStructure()
    cls1 = ontology.add_classification()
    assert cls1.uid == 1
    assert cls1.feature_node_hash
    cls2 = ontology.add_classification()
    assert cls2.uid == 2
    assert cls2.feature_node_hash

    feature_node_hash = short_uuid_str()
    cls3 = ontology.add_classification(uid=7, feature_node_hash=feature_node_hash)
    assert cls3.uid == 7
    assert cls3.feature_node_hash == feature_node_hash

    assert cls1.feature_node_hash != cls2.feature_node_hash

    assert len(ontology.classifications) == 3


def test_add_classification_duplicate_values():
    ontology = encord.objects.OntologyStructure()
    obj1 = ontology.add_classification(1, "12345678")
    with pytest.raises(ValueError):
        _ = ontology.add_classification(1)
    with pytest.raises(ValueError):
        _ = ontology.add_classification(2, feature_node_hash=obj1.feature_node_hash)

    assert len(ontology.classifications) == 1


def test_add_object():
    ontology = encord.objects.OntologyStructure()
    obj1 = ontology.add_object("Apple", Shape.BOUNDING_BOX)
    assert obj1.uid == 1
    assert obj1.color
    assert obj1.feature_node_hash
    obj2 = ontology.add_object("Orange", Shape.BOUNDING_BOX)
    assert obj2.uid == 2
    assert obj2.color
    assert obj2.feature_node_hash

    feature_node_hash = short_uuid_str()
    obj3 = ontology.add_object("Lemon", Shape.POLYLINE, 7, feature_node_hash=feature_node_hash)

    assert obj3.uid == 7
    assert obj3.color
    assert obj3.feature_node_hash == feature_node_hash

    assert obj1.feature_node_hash != obj2.feature_node_hash
    assert len(set([obj.color for obj in [obj1, obj2, obj3]])) == 3  # all the colors are different: auto-assigned

    assert len(ontology.objects) == 3


def test_add_object_duplicate_values():
    ontology = encord.objects.OntologyStructure()
    obj1 = ontology.add_object("Apple", Shape.BOUNDING_BOX, 1, "#000000", "12345678")
    with pytest.raises(ValueError):
        _ = ontology.add_object("Orange", Shape.BOUNDING_BOX, 1)
    with pytest.raises(ValueError):
        _ = ontology.add_object("Orange", Shape.BOUNDING_BOX, feature_node_hash=obj1.feature_node_hash)

    assert len(ontology.objects) == 1


def test_add_object_nested_classifications():
    ontology = encord.objects.OntologyStructure()
    obj1 = ontology.add_object("Apple", Shape.BOUNDING_BOX, 1, "#000000", "12345678")

    stripes = obj1.add_attribute(encord.objects.TextAttribute, "Stripes")
    assert stripes.uid == [1, 1]
    assert stripes.feature_node_hash
    assert isinstance(stripes, encord.objects.TextAttribute)
    assert not stripes.required
    assert not stripes.dynamic

    ripeness = obj1.add_attribute(
        encord.objects.RadioAttribute,
        "Ripeness",
        local_uid=7,
        feature_node_hash="12345678",
        required=True,
        dynamic=True,
    )
    assert ripeness.uid == [1, 7]
    assert ripeness.feature_node_hash
    assert isinstance(ripeness, encord.objects.RadioAttribute)
    assert ripeness.required
    assert ripeness.dynamic

    damage = obj1.add_attribute(encord.objects.ChecklistAttribute, "Damage", required=True, dynamic=True)
    assert damage.uid == [1, 8]
    assert damage.feature_node_hash
    assert isinstance(damage, encord.objects.ChecklistAttribute)
    assert damage.required
    assert damage.dynamic


def test_add_object_nested_classifications_duplicate_values():
    ontology = encord.objects.OntologyStructure()
    obj1 = ontology.add_object("Apple", Shape.BOUNDING_BOX, 1, "#000000", "12345678")

    attr1 = obj1.add_attribute(encord.objects.TextAttribute, "Stripes")
    with pytest.raises(ValueError):
        obj1.add_attribute(encord.objects.ChecklistAttribute, "Stars, I guess?", local_uid=1)
    with pytest.raises(ValueError):
        obj1.add_attribute(
            encord.objects.ChecklistAttribute, "Stars, I guess?", feature_node_hash=attr1.feature_node_hash
        )


def test_add_classification_attribute():
    ontology = encord.objects.OntologyStructure()
    cls1 = ontology.add_classification()

    clouds = cls1.add_attribute(encord.objects.RadioAttribute, "Cloud cover")
    assert clouds.uid == [1, 1]
    assert clouds.feature_node_hash
    assert isinstance(clouds, encord.objects.RadioAttribute)
    assert not clouds.required

    with pytest.raises(ValueError):  # only one root attribute per classification is allowed
        cls1.add_attribute(encord.objects.TextAttribute, "metadata")


def test_build_checkbox_options():
    ontology = encord.objects.OntologyStructure()
    cls1 = ontology.add_classification()

    clouds = cls1.add_attribute(encord.objects.ChecklistAttribute, "Cloud cover")
    one = clouds.add_option("Type One")
    assert one.value == "type_one"
    two = clouds.add_option("Type Two", value="two", local_uid=6)
    assert two.value == "two"
    three = clouds.add_option("Type Three", feature_node_hash="12345678")

    assert len(clouds.options) == 3
    assert len({opt.feature_node_hash for opt in clouds.options}) == 3  # all different
    assert three.uid == [1, 1, 7]

    assert isinstance(one, encord.objects.FlatOption)


def test_build_nested_options():
    ontology = encord.objects.OntologyStructure()
    cls1 = ontology.add_classification()

    clouds = cls1.add_attribute(encord.objects.RadioAttribute, "Cloud cover")
    one = clouds.add_option("Type One")
    two = clouds.add_option("Type Two", value="two", local_uid=6)

    assert isinstance(one, encord.objects.NestableOption)
    assert isinstance(two, encord.objects.NestableOption)

    detail1 = one.add_nested_attribute(encord.objects.RadioAttribute, "detail one")
    one.add_nested_attribute(encord.objects.TextAttribute, "detail two")

    detail1value1 = detail1.add_option("value 1")
    assert isinstance(detail1value1, encord.objects.NestableOption)
    assert detail1value1.uid == [1, 1, 1, 1, 1]  # five levels: root, 'clouds', 'type one', 'detail one', 'value 1'


def build_expected_ontology():
    ontology = encord.objects.OntologyStructure()

    _ = ontology.add_object(
        name="Eye",
        color="#D33115",
        shape=encord.objects.Shape.BOUNDING_BOX,
        feature_node_hash="a55abbeb",
    )
    nose = ontology.add_object(
        name="Nose",
        color="#E27300",
        shape=encord.objects.Shape.POLYGON,
        feature_node_hash="86648f32",
    )
    nose_detail = nose.add_attribute(
        encord.objects.ChecklistAttribute,
        feature_node_hash="1e3e5cad",
        name="Additional details about the nose",
        required=True,
    )
    nose_detail.add_option(feature_node_hash="2bc17c88", label="Is it a cute nose?")
    nose_detail.add_option(feature_node_hash="86eaa4f2", label="Is it a wet nose? ")
    example = ontology.add_object(
        name="Example",
        color="#FE9200",
        shape=encord.objects.Shape.POLYLINE,
        feature_node_hash="6eeba59b",
    )
    radio = example.add_attribute(
        encord.objects.RadioAttribute, feature_node_hash="cabfedb5", name="Radio with options"
    )
    nested = radio.add_option(feature_node_hash="5d102ce6", label="Nested Option")
    _ = nested.add_nested_attribute(encord.objects.RadioAttribute, feature_node_hash="59204845", name="Leaf")
    cls = ontology.add_classification(feature_node_hash="a39d81c0")
    cat_standing = cls.add_attribute(
        encord.objects.RadioAttribute,
        feature_node_hash="a6136d14",
        name="Is the cat standing?",
        required=True,
    )
    cat_standing.add_option(feature_node_hash="a3aeb48d", label="Yes")
    cat_standing.add_option(feature_node_hash="d0a4b373", label="No")

    assert ontology.to_dict() == EXPECTED_ONTOLOGY.to_dict()


def test_ontology_getters():
    # Object
    assert EXPECTED_ONTOLOGY.get_child_by_hash(OBJECT_1.feature_node_hash) == OBJECT_1
    assert EXPECTED_ONTOLOGY.get_child_by_hash(OBJECT_1.feature_node_hash, encord.objects.Object)
    with pytest.raises(TypeError):
        EXPECTED_ONTOLOGY.get_child_by_hash(OBJECT_1.feature_node_hash, encord.objects.TextAttribute)

    assert EXPECTED_ONTOLOGY.get_children_by_title(OBJECT_1.name) == [OBJECT_1]
    assert EXPECTED_ONTOLOGY.get_children_by_title(OBJECT_1.name, encord.objects.Object) == [OBJECT_1]
    assert EXPECTED_ONTOLOGY.get_children_by_title(OBJECT_1.name, encord.objects.Classification) == []

    # Option
    assert EXPECTED_ONTOLOGY.get_child_by_hash(FLAT_OPTION_1.feature_node_hash) == FLAT_OPTION_1
    assert EXPECTED_ONTOLOGY.get_child_by_hash(FLAT_OPTION_1.feature_node_hash, encord.objects.FlatOption)
    with pytest.raises(TypeError):
        EXPECTED_ONTOLOGY.get_child_by_hash(FLAT_OPTION_1.feature_node_hash, encord.objects.TextAttribute)

    assert EXPECTED_ONTOLOGY.get_children_by_title(FLAT_OPTION_1.label) == [FLAT_OPTION_1]
    assert EXPECTED_ONTOLOGY.get_children_by_title(FLAT_OPTION_1.label, encord.objects.FlatOption) == [FLAT_OPTION_1]
    assert EXPECTED_ONTOLOGY.get_children_by_title(FLAT_OPTION_1.label, encord.objects.Classification) == []

    # Attribute
    assert EXPECTED_ONTOLOGY.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash) == CHECKLIST_ATTRIBUTE
    assert EXPECTED_ONTOLOGY.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash, encord.objects.ChecklistAttribute)
    with pytest.raises(TypeError):
        EXPECTED_ONTOLOGY.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash, encord.objects.TextAttribute)

    assert EXPECTED_ONTOLOGY.get_children_by_title(CHECKLIST_ATTRIBUTE.name) == [CHECKLIST_ATTRIBUTE]
    assert EXPECTED_ONTOLOGY.get_children_by_title(CHECKLIST_ATTRIBUTE.name, encord.objects.ChecklistAttribute) == [
        CHECKLIST_ATTRIBUTE
    ]
    assert EXPECTED_ONTOLOGY.get_children_by_title(CHECKLIST_ATTRIBUTE.name, encord.objects.Classification) == []

    # Classification
    assert EXPECTED_ONTOLOGY.get_child_by_hash(CLASSIFICATION_1.feature_node_hash) == CLASSIFICATION_1
    assert EXPECTED_ONTOLOGY.get_child_by_hash(CLASSIFICATION_1.feature_node_hash, encord.objects.Classification)
    with pytest.raises(TypeError):
        EXPECTED_ONTOLOGY.get_child_by_hash(CLASSIFICATION_1.feature_node_hash, encord.objects.TextAttribute)
    # NOTE: getting by name does not work. The classification has no name, just its attribute


def test_object_getters():
    # Option
    assert OBJECT_2.get_child_by_hash(FLAT_OPTION_1.feature_node_hash) == FLAT_OPTION_1
    assert OBJECT_2.get_child_by_hash(FLAT_OPTION_1.feature_node_hash, encord.objects.FlatOption)
    with pytest.raises(TypeError):
        OBJECT_2.get_child_by_hash(FLAT_OPTION_1.feature_node_hash, encord.objects.TextAttribute)

    assert OBJECT_2.get_children_by_title(FLAT_OPTION_1.label) == [FLAT_OPTION_1]
    assert OBJECT_2.get_children_by_title(FLAT_OPTION_1.label, encord.objects.FlatOption) == [FLAT_OPTION_1]
    assert OBJECT_2.get_children_by_title(FLAT_OPTION_1.label, encord.objects.ChecklistAttribute) == []

    # Attribute
    assert OBJECT_2.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash) == CHECKLIST_ATTRIBUTE
    assert OBJECT_2.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash, encord.objects.ChecklistAttribute)
    with pytest.raises(TypeError):
        OBJECT_2.get_child_by_hash(CHECKLIST_ATTRIBUTE.feature_node_hash, encord.objects.TextAttribute)

    assert OBJECT_2.get_children_by_title(CHECKLIST_ATTRIBUTE.name) == [CHECKLIST_ATTRIBUTE]
    assert OBJECT_2.get_children_by_title(CHECKLIST_ATTRIBUTE.name, encord.objects.ChecklistAttribute) == [
        CHECKLIST_ATTRIBUTE
    ]
    assert OBJECT_2.get_children_by_title(CHECKLIST_ATTRIBUTE.name, encord.objects.TextAttribute) == []


def test_classification_getters():
    # Option
    assert CLASSIFICATION_1.get_child_by_hash(NESTABLE_OPTION_2.feature_node_hash) == NESTABLE_OPTION_2
    assert CLASSIFICATION_1.get_child_by_hash(NESTABLE_OPTION_2.feature_node_hash, encord.objects.NestableOption)
    with pytest.raises(TypeError):
        CLASSIFICATION_1.get_child_by_hash(NESTABLE_OPTION_2.feature_node_hash, encord.objects.TextAttribute)

    assert CLASSIFICATION_1.get_children_by_title(NESTABLE_OPTION_2.label) == [NESTABLE_OPTION_2]
    assert CLASSIFICATION_1.get_children_by_title(NESTABLE_OPTION_2.label, encord.objects.NestableOption) == [
        NESTABLE_OPTION_2
    ]
    assert CLASSIFICATION_1.get_children_by_title(NESTABLE_OPTION_2.label, encord.objects.ChecklistAttribute) == []

    # Attribute
    assert CLASSIFICATION_1.get_child_by_hash(RADIO_ATTRIBUTE_3.feature_node_hash) == RADIO_ATTRIBUTE_3
    assert CLASSIFICATION_1.get_child_by_hash(RADIO_ATTRIBUTE_3.feature_node_hash, encord.objects.RadioAttribute)
    with pytest.raises(TypeError):
        CLASSIFICATION_1.get_child_by_hash(RADIO_ATTRIBUTE_3.feature_node_hash, encord.objects.TextAttribute)

    assert CLASSIFICATION_1.get_children_by_title(RADIO_ATTRIBUTE_3.name) == [RADIO_ATTRIBUTE_3]
    assert CLASSIFICATION_1.get_children_by_title(RADIO_ATTRIBUTE_3.name, encord.objects.RadioAttribute) == [
        RADIO_ATTRIBUTE_3
    ]
    assert CLASSIFICATION_1.get_children_by_title(RADIO_ATTRIBUTE_3.name, encord.objects.TextAttribute) == []


def test_attribute_getters():
    # Option
    assert RADIO_ATTRIBUTE_2.get_child_by_hash(NESTABLE_OPTION_3.feature_node_hash) == NESTABLE_OPTION_3
    assert RADIO_ATTRIBUTE_2.get_child_by_hash(NESTABLE_OPTION_3.feature_node_hash, encord.objects.NestableOption)
    with pytest.raises(TypeError):
        RADIO_ATTRIBUTE_2.get_child_by_hash(NESTABLE_OPTION_3.feature_node_hash, encord.objects.TextAttribute)

    assert RADIO_ATTRIBUTE_2.get_children_by_title(NESTABLE_OPTION_3.label) == [NESTABLE_OPTION_3]
    assert RADIO_ATTRIBUTE_2.get_children_by_title(NESTABLE_OPTION_3.label, encord.objects.NestableOption) == [
        NESTABLE_OPTION_3
    ]
    assert RADIO_ATTRIBUTE_2.get_children_by_title(NESTABLE_OPTION_3.label, encord.objects.ChecklistAttribute) == []

    # Attribute
    assert RADIO_ATTRIBUTE_2.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash) == RADIO_ATTRIBUTE_1
    assert RADIO_ATTRIBUTE_2.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash, encord.objects.RadioAttribute)
    with pytest.raises(TypeError):
        RADIO_ATTRIBUTE_2.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash, encord.objects.TextAttribute)

    assert RADIO_ATTRIBUTE_2.get_children_by_title(RADIO_ATTRIBUTE_1.name) == [RADIO_ATTRIBUTE_1]
    assert RADIO_ATTRIBUTE_2.get_children_by_title(RADIO_ATTRIBUTE_1.name, encord.objects.RadioAttribute) == [
        RADIO_ATTRIBUTE_1
    ]
    assert RADIO_ATTRIBUTE_2.get_children_by_title(RADIO_ATTRIBUTE_1.name, encord.objects.TextAttribute) == []


def test_option_getters():
    # Attribute
    assert NESTABLE_OPTION_3.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash) == RADIO_ATTRIBUTE_1
    assert NESTABLE_OPTION_3.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash, encord.objects.RadioAttribute)
    with pytest.raises(TypeError):
        NESTABLE_OPTION_3.get_child_by_hash(RADIO_ATTRIBUTE_1.feature_node_hash, encord.objects.TextAttribute)

    assert NESTABLE_OPTION_3.get_children_by_title(RADIO_ATTRIBUTE_1.name) == [RADIO_ATTRIBUTE_1]
    assert NESTABLE_OPTION_3.get_children_by_title(RADIO_ATTRIBUTE_1.name, encord.objects.RadioAttribute) == [
        RADIO_ATTRIBUTE_1
    ]
    assert NESTABLE_OPTION_3.get_children_by_title(RADIO_ATTRIBUTE_1.name, encord.objects.TextAttribute) == []



from abc import ABC, abstractmethod
from typing import Any, Dict, Type, TypeVar

T = TypeVar("T", bound="BaseDTOInterface")


class BaseDTOInterface(ABC):
    @classmethod
    @abstractmethod
    def from_dict(cls: Type[T], d: Dict[str, Any]) -> T:
        pass

    @abstractmethod
    def to_dict(self, by_alias=True, exclude_none=True) -> Dict[str, Any]:
        pass



from collections import OrderedDict

from encord.orm import base_orm


class LabelingAlgorithm(base_orm.BaseORM):
    """Labeling algorithm base ORM.

    ORM:

    algorithm_name,
    algorithm_params

    """

    DB_FIELDS = OrderedDict(
        [
            ("algorithm_name", str),
            ("algorithm_parameters", dict),  # Algorithm params
        ]
    )


class ObjectInterpolationParams(base_orm.BaseORM):
    """Labeling algorithm parameters for interpolation algorithm

    ORM:

    key_frames,
    objects_to_interpolate

    """

    DB_FIELDS = OrderedDict(
        [
            ("key_frames", dict),
            ("objects_to_interpolate", list),
        ]
    )


class BoundingBoxFittingParams(base_orm.BaseORM):
    """Labeling algorithm parameters for bounding box fitting algorithm

    ORM:

    labels,
    video

    """

    DB_FIELDS = OrderedDict(
        [
            ("labels", dict),
            ("video", dict),
        ]
    )



labels = {
    "label_hash": "003959e1-a63f-4e08-9f3c-916d1e807ed6",
    "branch_name": "main",
    "created_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "last_edited_at": "Thu, 09 Feb 2023 14:12:03 UTC",
    "data_hash": "aaa6bc82-9f89-4545-adbb-f271bf28cf99",
    "annotation_task_status": "QUEUED",
    "is_shadow_data": False,
    "dataset_hash": "b0d93919-a5e8-4418-8dd5-2c51e3977de8",
    "dataset_title": "Dataset with 2 frame video",
    "data_title": "two-frame-video.mp4",
    "data_type": "video",
    "data_units": {
        "cd57cf5c-2541-4a46-a836-444540ee987a": {
            "data_hash": "cd57cf5c-2541-4a46-a836-444540ee987a",
            "data_title": "two-frame-video.mp4",
            "data_type": "video/mp4",
            "data_sequence": 0,
            "labels": {
                "0": {
                    "objects": [
                        {
                            "name": "box",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "box",
                            "createdAt": "Tue, 17 Jan 2023 17:23:10 UTC",
                            "createdBy": "denis@cord.tech",
                            "confidence": 1,
                            "objectHash": "0gCzNuLS",
                            "lastEditedAt": "Wed, 18 Jan 2023 17:23:24 UTC",
                            "featureHash": "LJNV5PVK",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.1291, "w": 0.1296, "x": 0.4287, "y": 0.2615},
                        },
                        {
                            "name": "box 2",
                            "color": "#E27300",
                            "shape": "bounding_box",
                            "value": "box_2",
                            "createdAt": "Tue, 17 Jan 2023 17:23:36 UTC",
                            "createdBy": "denis@cord.tech",
                            "lastEditedAt": "Wed, 18 Jan 2023 17:23:26 UTC",
                            "confidence": 1,
                            "objectHash": "Wpc+TmG+",
                            "featureHash": "GaAKWvvg",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.0926, "w": 0.0929, "x": 0.5637, "y": 0.5477},
                        },
                    ],
                    "classifications": [],
                },
                "1": {
                    "objects": [
                        {
                            "name": "box",
                            "color": "#D33115",
                            "shape": "bounding_box",
                            "value": "box",
                            "createdAt": "Tue, 17 Jan 2023 17:23:21 UTC",
                            "createdBy": "denis@cord.tech",
                            "isDeleted": False,
                            "confidence": 1,
                            "objectHash": "0gCzNuLS",
                            "featureHash": "LJNV5PVK",
                            "lastEditedAt": "Tue, 17 Jan 2023 17:23:22 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.1291, "w": 0.1296, "x": 0.4449, "y": 0.2489},
                        },
                        {
                            "name": "box 2",
                            "color": "#E27300",
                            "shape": "bounding_box",
                            "value": "box_2",
                            "createdAt": "Tue, 17 Jan 2023 17:23:38 UTC",
                            "createdBy": "denis@cord.tech",
                            "isDeleted": False,
                            "confidence": 1,
                            "objectHash": "Wpc+TmG+",
                            "featureHash": "GaAKWvvg",
                            "lastEditedAt": "Tue, 17 Jan 2023 17:23:40 UTC",
                            "lastEditedBy": "denis@cord.tech",
                            "manualAnnotation": True,
                            "boundingBox": {"h": 0.0926, "w": 0.0929, "x": 0.5799, "y": 0.5505},
                        },
                    ],
                    "classifications": [],
                },
            },
            "data_fps": 25.0,
            "data_duration": 0.08,
            "data_link": "cord-videos-dev/lFW59RQ9jcT4vHZeG14m8QWJKug1/cd57cf5c-2541-4a46-a836-444540ee987a",
            "width": 1200,
            "height": 924,
        }
    },
    "object_answers": {
        "0gCzNuLS": {"objectHash": "0gCzNuLS", "classifications": []},
        "Wpc+TmG+": {"objectHash": "Wpc+TmG+", "classifications": []},
    },
    "classification_answers": {},
    "object_actions": {
        "0gCzNuLS": {
            "objectHash": "0gCzNuLS",
            "actions": [
                {
                    "name": "dynamic text",
                    "range": [[0, 0]],
                    "value": "dynamic_text",
                    "answers": "dynamic text",
                    "dynamic": True,
                    "trackHash": "fbb97dda-1e66-48f9-b749-af2f83dab9fc",
                    "featureHash": "5RGh3YIU",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
                {
                    "name": "dynamic classification",
                    "range": [[0, 0]],
                    "value": "dynamic_classification",
                    "answers": [{"name": "classification 1", "value": "classification_1", "featureHash": "aOQGJvce"}],
                    "dynamic": True,
                    "trackHash": "9b3e2e91-3c4a-45dd-a142-b9e53642f1c3",
                    "featureHash": "ZJoiBceM",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
                {
                    "name": "dynamic radio",
                    "range": [[0, 0]],
                    "value": "dynamic_radio",
                    "answers": [{"name": "radio 1", "value": "radio_1", "featureHash": "ps0dwXfI"}],
                    "dynamic": True,
                    "trackHash": "6c2e03f4-1450-48a3-8f9f-6761af3006a3",
                    "featureHash": "OMpf1zPf",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
                {
                    "name": "dynamic text",
                    "range": [[1, 1]],
                    "value": "dynamic_text",
                    "answers": "dynamic text next frame",
                    "dynamic": True,
                    "trackHash": "e138b677-03d5-48f6-a4d7-806964e60339",
                    "featureHash": "5RGh3YIU",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
                {
                    "name": "dynamic radio",
                    "range": [[1, 1]],
                    "value": "dynamic_radio",
                    "answers": [{"name": "radio 2", "value": "radio_2", "featureHash": "IwCHSdeL"}],
                    "dynamic": True,
                    "trackHash": "05dc3b48-19fc-402b-8323-301ca52f410c",
                    "featureHash": "OMpf1zPf",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
                {
                    "name": "dynamic classification",
                    "range": [[1, 1]],
                    "value": "dynamic_classification",
                    "answers": [
                        {"name": "classification 2", "value": "classification_2", "featureHash": "tyjScPrk"},
                        {"name": "classification 1", "value": "classification_1", "featureHash": "aOQGJvce"},
                    ],
                    "dynamic": True,
                    "trackHash": "69b2a3d0-3db4-41d9-ba2a-d200aecd3d23",
                    "featureHash": "ZJoiBceM",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                },
            ],
        },
        "Wpc+TmG+": {
            "objectHash": "Wpc+TmG+",
            "actions": [
                {
                    "name": "radio 2",
                    "range": [[1, 1]],
                    "value": "radio_2",
                    "answers": [{"name": "radio 2 2", "value": "radio_2_2", "featureHash": "J/I/karX"}],
                    "dynamic": True,
                    "trackHash": "2e21dbab-e3b3-4a39-8fd2-def4c7745b4f",
                    "featureHash": "TWPByH4G",
                    "shouldPropagate": False,
                    "manualAnnotation": True,
                }
            ],
        },
    },
    "label_status": "LABEL_IN_PROGRESS",
}
